theory GCD_Fred_CAS
  imports   Isabelle_C_AutoCorres.AutoCorres_Wrapper
begin

text\<open>This is Fred's GCD example\<close>


section \<open>Setup of Representation Function \<open>C\<^sub>e\<^sub>x\<^sub>p\<^sub>r\<close>\<close>


setup \<open>C_Module.C_Term.map_expression
        (fn expr => fn _ => fn _ => 
          case expr of C_Ast.CVar0 (C_Ast.Ident0 (_, x, _), _) =>
                         Free (C_Grammar_Rule_Lib.ident_decode x, dummyT)
                     | s => Free (\<^make_string> s, dummyT))\<close>



section\<open>Background\<close>

text\<open>No further background other than the HOL-library necessary; in particular, 
     this program theory resides on the definition of @{term "GCD.gcd"}.\<close>



section\<open>The Program\<close>

declare [[AutoCorres]]

text \<open>Setup of AutoCorres for semantically representing this C element:\<close>


definition "pgcd_inv (a::int) b m n \<equiv> a > 0 \<and> b > 0 \<and>  
                                      m < INT_MAX \<and> n < INT_MAX \<and> 
                                      a \<le> m \<and> b \<le> n \<and>
                                      (\<forall> g. g = gcd a b \<longrightarrow> g = gcd m n)"



ML\<open>

fun definition (((decl, spec), prems), params) =
  #2 oo Specification.definition_cmd decl params prems spec;

fun definition2 (decl, spec, prems, params) = #2 oo Specification.definition' decl params prems spec



val X = (Scan.option Parse_Spec.constdecl -- (Parse_Spec.opt_thm_name ":" -- Parse.prop) --
        Parse_Spec.if_assumes -- Parse.for_fixes) 
        : ((((binding * string option * mixfix) option * (Attrib.binding * string)) * string list) *
      (binding * string option * mixfix) list) parser;

C_Inner_Syntax.local_command''
      ("definition", \<^here>, \<^here>, \<^here>)
      (Scan.option Parse_Spec.constdecl -- (Parse_Spec.opt_thm_name ":" -- Parse.prop) --
        Parse_Spec.if_assumes -- Parse.for_fixes)
      definition;

C_Inner_Syntax.local_command';
C_Inner_Syntax.command0;
C_Inner_Syntax.command00;
C_Inner_Syntax.command_no_range;
C_Inner_Syntax.command_no_range';
C_Inner_Syntax.command_no_range';
\<close>


ML\<open>val H : Input.source -> Inttab.key option -> Context.generic -> Context.generic =
              C_Inner_Toplevel.generic_theory oo C_Inner_Isar_Cmd.setup\<close>

ML\<open>
fun  mk_undefined (@{typ "unit"}) = Const (\<^const_name>\<open>Product_Type.Unity\<close>, \<^typ>\<open>unit\<close>)
    |mk_undefined t               = Const (\<^const_name>\<open>HOL.undefined\<close>, t)

fun meta_eq_const T = Const (\<^const_name>\<open>Pure.eq\<close>, T --> T --> propT);

fun mk_meta_eq (t, u) = meta_eq_const (fastype_of t) $ t $ u;

fun   mk_pat_tupleabs [] t = t
    | mk_pat_tupleabs [(s,ty)] t = absfree(s,ty)(t)
    | mk_pat_tupleabs ((s,ty)::R) t = HOLogic.mk_case_prod(absfree(s,ty)(mk_pat_tupleabs R t));


\<close>

ML\<open>


   fun define_cond binding f_sty transform_old  cond_suffix params read_cond (ctxt:local_theory) = 
       let val params' = map (fn(b, ty) => (Binding.name_of b,ty)) params
           val src' = case transform_old (read_cond ctxt) of 
                        Abs(nn, sty_pre, term) => mk_pat_tupleabs params' (Abs(nn,sty_pre,term))
                      | _ => error ("define abstraction for result" ^ Position.here \<^here>)
           val bdg = Binding.suffix_name cond_suffix binding
           val bdg_ty = HOLogic.mk_tupleT(map (#2) params) --> f_sty HOLogic.boolT
           val eq =  mk_meta_eq(Free(Binding.name_of bdg, bdg_ty),src')
           val args = (SOME(bdg,NONE,NoSyn), (Binding.empty_atts,eq),[],[]) 
       in  definition2 args true ctxt end

   fun define_precond binding sty =
       define_cond binding (fn boolT => sty --> boolT) I  "_requires" 

   fun define_postcond binding rty sty =
       define_cond binding (fn boolT => sty --> sty --> rty --> boolT) I "_ensures" 

 
 \<close>

ML\<open>
local 
val CMD = ref(K (K I):LALR_Table.state -> C_Env.env_lang -> Context.generic -> Context.generic);
in
fun command_general cmd = let val _ = (CMD:=cmd) in
         (H
         \<open>(fn ((tb:LALR_Table.state, (X : C_Grammar_Rule.svalue0, pos1, pos2)) :: Z) =>
              (fn stck (* Polytypic - depends on the navigator , for example for +@: 
                          : 'a C_Stack.stack_elem *) => 
                 fn cenv: C_Env.env_lang =>
                tap (fn ctxt : Context.generic => !CMD tb cenv ctxt))
           | _ => fn _ => fn _ => I )\<close>) 
         end
         ;  
end
\<close>

text\<open>@{ML_type "C_Ast.CTypeSpec C_Stack.stack_elem"}\<close>

ML\<open>
val Highlight = (H
         \<open>(fn ((tb:LALR_Table.state, (X : C_Grammar_Rule.svalue0, pos1, pos2)) :: Z) =>
              (fn stck (* Polytypic - depends on the navigator , for example 
                          for +@:  C_Ast.CTypeSpec C_Stack.stack_elem 
                          for ++@: C_Ast.CDeclrR C_Stack.stack_elem*)
                           => 
                 fn cenv: C_Env.env_lang =>
                tap (fn ctxt : Context.generic => (          
                      Position.reports_text [((Position.range (pos1, pos2)
                                               |> Position.range_position, Markup.intensify), "")])))
           | _ => fn _ => fn _ => I )\<close>) 
         ;  
(*
fun command' cmd scan f =
       C_Annotation.command' cmd "" (bind_export scan f)

fun command'' cmd A B C =
    command' cmd
           (C_Token.syntax' (Parse.token Parse.term))
           (fn src => fn _ => fn ctxt =>
             ML_Context.exec
               (tap (fn _ => Syntax.read_term (Context.proof_of ctxt)
                                              (Token.inner_syntax_of src)))
               ctxt)



val _ = command' : string * Position.T ->
      (binding option * string) C_Parse.parser ->
        (binding option * string -> C_Env.env_propagation_reduce -> Context.generic -> Context.generic)
          -> theory -> theory
*)

val _ = Theory.setup   (   C_Inner_Syntax.command_no_range Highlight  ("highlight2", \<^here>, \<^here>) )

(*
                        #> command' ("ensures2", \<^here>)   scan_lab_inner_syntax ensures)
                       
*)
\<close>







C\<open>

int x;

int pgcd(int m, int n) {
  //@ definition \<open>pgcd_requires m n = (\<lambda> \<sigma>.  m>0 \<and> n > 0 \<and> m \<le> UINT_MAX  \<and> n \<le> UINT_MAX)\<close>
  //@ +@ requires pre1 : \<open>\<^C>\<^sub>e\<^sub>x\<^sub>p\<^sub>r\<open>m\<close> > 0 \<and> \<^C>\<^sub>e\<^sub>x\<^sub>p\<^sub>r\<open>n\<close>  > 0 \<and>  \<^C>\<^sub>e\<^sub>x\<^sub>p\<^sub>r\<open>m\<close> \<le> UINT_MAX  \<and> \<^C>\<^sub>e\<^sub>x\<^sub>p\<^sub>r\<open>n\<close> \<le> UINT_MAX\<close>
  // the clause should be as above ...
  //@ definition "pgcd_ensures m n = (\<lambda>res \<sigma>.  res = gcd m n)"
  //@ +@ ensures post1 :  \<open>\<lambda>res. res = gcd \<^C>\<^sub>e\<^sub>x\<^sub>p\<^sub>r\<open>m\<close>  \<^C>\<^sub>e\<^sub>x\<^sub>p\<^sub>r\<open>n\<close> \<close>
  // the clause should be as above ...
  //@ ++@ highlight2

  int a = m;
  int b = n;

  while (a != b) {
  //@ definition "pgcd_inv\<^sub>1    m n = (\<lambda>(a, b) _. pgcd_inv a b m n)"
  //@ definition "pgcd_mesure\<^sub>1 m n = (\<lambda>((a, b),_) . nat a + nat b)"
  //@ INVARIANT : "\<lambda>(a, x) s. pgcd_inv a b \<^C>\<^sub>e\<^sub>x\<^sub>p\<^sub>r\<open>m\<close>  \<^C>\<^sub>e\<^sub>x\<^sub>p\<^sub>r\<open>n\<close>"
  //@ invariant  \<comment> \<open>inner\<close> \<open>\<lambda>(a, y) s. pgcd_inv a b \<^C>\<^sub>e\<^sub>x\<^sub>p\<^sub>r\<open>m\<close>  \<^C>\<^sub>e\<^sub>x\<^sub>p\<^sub>r\<open>n\<close> \<close>
  //@ measure    \<comment> \<open>inner\<close> \<open>\<lambda>((a, z),s) . nat a + nat b\<close>
    if (a < b) {
      b = b - a;
    } else {
      a = a - b;
    }
  }
  return a;
}

//@ install_autocorres pgcd [ ts_rules = nondet, unsigned_word_abs = pgcd ]

\<close>

ML\<open>Inttab.lookup;
C_Module'.Annotation_Data.get;
C_Module'.Annotation_Data.put;
C_Module'.Annotation_Data.map;
 \<close>

ML\<open>
val get_anno_data = C_Module'.Annotation_Data.get ;
val bbb = get_anno_data (Context.Proof @{context})
\<close>


ML\<open>
val get_anno_data = C_Module'.Annotation_Data.get ;
val blocks = Inttab.keys o get_anno_data;

val get_block  =  (the oo Inttab.lookup) o  get_anno_data; (* polymorphism *)
fun swapC f a b = f b a;
fun filter_tag_of_block (k,tag) ad = ad |> (swapC get_block) k
                                        |> filter (fn X => tag = #tag X)
val get_invariants = filter_tag_of_block (1,"invariant")
val get_measures   = filter_tag_of_block (1,"measure")
val get_requires   = filter_tag_of_block (1,"requires")
val get_ensures    = filter_tag_of_block (1,"ensures")

val a = get_requires (Context.Proof @{context})
\<close>

ML\<open>
val [(n,[A,B,C,D])] = Inttab.dest bbb;
val tag1 = (#tag A);
val t1 = Syntax.read_term @{context} (#content A);
val t2 = Syntax.read_term @{context} (#content B);
val Env1 = (#c_env A)
val Env2 = (#c_env B)

fun get_symtab       (env : C_Env.env_lang) = (#idents o snd o hd (* ? *) o #scopes) env
fun get_symtab_ids   (env : C_Env.env_lang) = Symtab.keys(get_symtab env)
fun get_symtab_global_ids(env : C_Env.env_lang) =
                     map_filter (fn(x,y) => if (#global o #3) y then SOME x else NONE)
                                (Symtab.dest(get_symtab env))

fun symtab_lookup_id (env : C_Env.env_lang) (id:string) =
                            the(Symtab.lookup (get_symtab env) id)

val {  var_table 
     , scopes : (C_Ast.ident option * C_Env.var_table) list 
     , namesupply : int
     , stream_ignored : C_Antiquote.antiq C_Env.stream
     , env_directives : C_Env.env_directives } = (#c_env A)\<close>

ML\<open> \<close>

ML\<open>(*get_symtab_ids (#c_env A); *)

   get_symtab_global_ids (#c_env A);

   (* symtab_lookup_id (#c_env A) "n" *)
  val [C_Ast.CFunDeclr0 X] = #params(#3(symtab_lookup_id (#c_env A) "pgcd"));
  val (Y1,Y2,Y3) = X ;
  val C_Ast.Right Y4 = Y1;
\<close>

ML\<open>
val ([C_Ast.CDecl0 Y5,C_Ast.CDecl0 Y5'],false) =Y4;
val (Y6,Y6',Y6'') = Y5

\<close>

ML\<open>val [C_Ast.CTypeSpec0 Y7] = Y6
val C_Ast.CIntType0 (C_Ast.NodeInfo0 Y8) = Y7
\<close>


ML\<open>open  C_Ast\<close>
ML\<open>
fun  map_cTypeSpecifier f g (CAtomicType0 (_, a)) = f a
   | map_cTypeSpecifier f g (CBoolType0 a)  = f a
   | map_cTypeSpecifier f g (CCharType0 a)  = f a
   | map_cTypeSpecifier f g (CComplexType0 a) = f a
   | map_cTypeSpecifier f g (CDoubleType0 a) = f a
   | map_cTypeSpecifier f g (CEnumType0(_, a)) = f a
   | map_cTypeSpecifier f g (CFloatType0 a)  = f a
   | map_cTypeSpecifier f g (CInt128Type0 a)  = f a
   | map_cTypeSpecifier f g (CIntType0 a)  = f a
   | map_cTypeSpecifier f g (CLongType0 a)  = f a
   | map_cTypeSpecifier f g (CSUType0 (_, a)) = f a
   | map_cTypeSpecifier f g (CShortType0 a)  = f a
   | map_cTypeSpecifier f g (CSignedType0 a)  = f a
   | map_cTypeSpecifier f g (CTypeDef0 (_, a)) = f a
   | map_cTypeSpecifier f g (CTypeOfExpr0 (_, a)) = f a
   | map_cTypeSpecifier f g (CTypeOfType0 (_, a)) = f a
   | map_cTypeSpecifier f g (CUnsigType0 a)  = f a
   | map_cTypeSpecifier f g (CVoidType0 a)  = f a

\<close>

text\<open>regarde @{ML_file "../type_strengthen.ML"} \<close>
 

print_theorems
find_theorems name:GCD

find_theorems name:ensures

section\<open>The Correctness-Proof\<close>



theorem (in pgcd) pgcd'_correct:
    "\<lbrace> \<lambda>\<sigma>. m > 0 \<and> n > 0 \<and> m < INT_MAX \<and> n < INT_MAX \<rbrace> pgcd' m n \<lbrace> \<lambda>res \<sigma>. res = gcd m n \<rbrace>!"
  unfolding pgcd'_def
  text\<open>Turning precond into explicit assumption\<close>
  apply(rule validNF_assume_pre)
  text\<open>Annotating loop with invariant and measure\<close>
  apply(subst whileLoop_add_inv   [  where I = "\<lambda>(a, b) s.   pgcd_inv a b m n"
                                     and   M = "\<lambda>((a, b),s). nat a + nat b"])
  proof(wp, fold INT_MAX_def, auto simp: pgcd_inv_def)
     fix a b show " gcd a b = gcd m n \<Longrightarrow> gcd a (b - a) = gcd m n"
          by (metis (no_types, hide_lams) gcd.commute gcd_diff1)
  next
     fix a b show " gcd a b = gcd m n \<Longrightarrow> gcd (a - b) b = gcd m n"
          by (simp add: gcd_diff1)
  qed

end